/* dc_gui2 - a GTK+2 GUI for DCTC
 * Copyright (C) 2002 Eric Prevoteau
 *
 * public_hub_clist.c: Copyright (C) Eric Prevoteau <www@ac2i.tzo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
$Id: public_hub_clist.c,v 1.6 2003/06/01 08:03:23 eric Exp $
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>
#include <gnome.h>

#include "gtkcellrenderertextx.h"
#include "misc_gtk.h"
#include "public_hub_clist.h"
#include "seen_hub_clist.h"
#include "main.h"
#include "gui_define.h"
#include "macro.h"
#include "status.h"
#include "init_fnc.h"
#include "gtk_helper.h"
#include "sound.h"

/**********************************************************/
/* build hublist command line according to GUI parameters */
/**********************************************************/
static void start_hublist(void)
{
	GStringChunk *gsc=NULL;
	GPtrArray *gpa=NULL;
	int valid_len;
	int i;
	const char *cnxtype;
	char *hubcnxtype[3]={_("No proxy needed"), _("Use SOCKS parameters"), _("Use Web proxy")};
	GtkWidget *w;
	const char *t;

	/* update the HUBLIST shell variable to match the current config */
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget_by_widget_name(main_window,"hublist_radiobutton")))==TRUE)
	{
		unsetenv("HUBLIST");
	}
	else
	{
		t=gtk_entry_get_text(GTK_ENTRY(get_widget_by_widget_name(main_window,"hidden_planar_hublist")));
		if((t==NULL)||(strlen(t)==0))
		{
			unsetenv("HUBLIST");
		}
		else
		{
			setenv("HUBLIST",t,TRUE);
		}
	}

	gsc=g_string_chunk_new(128);
	gpa=g_ptr_array_new();

	g_ptr_array_add(gpa,g_string_chunk_insert(gsc,"hublist"));	/* argv[0] */

	cnxtype=gtk_entry_get_text(GTK_ENTRY(get_widget_by_widget_name(main_window,"hublist_cnxtype_entry")));
	for(i=0;i<3;i++)
	{
		if(!strcmp(cnxtype,hubcnxtype[i]))
			break;
	}

	valid_len=gpa->len;

	switch(i)
	{
		case 0:		/* no proxy needed */
						break;

		case 1:		/* use socks param */
						g_ptr_array_add(gpa,g_string_chunk_insert(gsc,"--socks"));
						{
							/* add socks proxy address */
							w=get_widget_by_widget_name(main_window,"socks_address_entry");
							if(w==NULL)
								break;
							
							t=gtk_entry_get_text(GTK_ENTRY(w));
							if((t==NULL)||(strlen(t)==0))
								break;

							g_ptr_array_add(gpa,g_string_chunk_insert(gsc,t));
						}
						{
							/* add socks proxy port */
							w=get_widget_by_widget_name(main_window,"socks_port_entry");
							if(w==NULL)
								break;
							
							t=gtk_entry_get_text(GTK_ENTRY(w));
							if((t==NULL)||(strlen(t)==0))
								break;

							g_ptr_array_add(gpa,g_string_chunk_insert(gsc,t));
							valid_len=gpa->len;
						}
						{
							/* add socks user ID (if exist) */
							w=get_widget_by_widget_name(main_window,"socks_userid_entry");
							if(w==NULL)
								break;
							
							t=gtk_entry_get_text(GTK_ENTRY(w));
							if((t==NULL)||(strlen(t)==0))
								break;

							g_ptr_array_add(gpa,g_string_chunk_insert(gsc,t));
							valid_len=gpa->len;
						}
						break;

		case 2:		/* use web proxy */
						g_ptr_array_add(gpa,g_string_chunk_insert(gsc,"--proxy"));
						{
							/* add web proxy address */
							w=get_widget_by_widget_name(main_window,"web_proxy_host_entry");
							if(w==NULL)
								break;
							
							t=gtk_entry_get_text(GTK_ENTRY(w));
							if((t==NULL)||(strlen(t)==0))
								break;

							g_ptr_array_add(gpa,g_string_chunk_insert(gsc,t));
						}
						{
							/* add socks proxy port */
							w=get_widget_by_widget_name(main_window,"web_proxy_port_entry");
							if(w==NULL)
								break;
							
							t=gtk_entry_get_text(GTK_ENTRY(w));
							if((t==NULL)||(strlen(t)==0))
								break;

							g_ptr_array_add(gpa,g_string_chunk_insert(gsc,t));
							valid_len=gpa->len;
						}
						break;

		default:		/* eh ??? */
						break;
	}

	/* discard any incomplete parameter */
	if(valid_len==gpa->len)
		g_ptr_array_add(gpa,NULL);		/* ends the array */
	else
		gpa->pdata[valid_len]=NULL;

	/* and start hublist */
	execvp(gpa->pdata[0],(char **)(gpa->pdata));
	
	/* on error, we leave */
	g_string_chunk_free(gsc);
	g_ptr_array_free(gpa,TRUE);
}

static guint fill_pubhub_handle_id=-1;			/* when !=-1, it is the gdk_input handle for the function updating the hub_public_clist */
static int fill_pubhub_pipe=-1;					/* pipe used to communicate with hub list (it is a read only pipe) */
static pid_t fill_pubhub_procpid=-1;			/* when !=-1, it is the pid of the hublist program */
static GtkTreeModel *fill_pubhub_hpc_gtm=NULL;	/* it is the GtkTreeModel of the "hub_public_clist" */
static GtkListStore *fill_pubhub_hpc_gls=NULL;	/* it is the GtkListStore of the "hub_public_clist" */
static GString *fill_pubhub_incoming_data=NULL;	/* string containing received data */

/**************************************************************************/
/* fill the clist named "hub_public_clist" with the list of existing hubs */
/**************************************************************************/
static void async_fill_pub_hub_clist(gpointer data, gint source, GdkInputCondition condition)
{
	int rd;
	gchar **row=NULL;
	GPtrArray *gpa=NULL;
	GStringChunk *gsc=NULL;
	int i;
	SORT_VARS( )

	char buf[1024+1];

	rd=read(source,buf,sizeof(buf)-1);
	if(rd==-1)
	{
		if((errno==EAGAIN)||(errno==EINTR))		/* come back later */
			return;
	}
	else if(rd!=0)			/* we have something */
	{
		buf[rd]='\0';
		fill_pubhub_incoming_data=g_string_append(fill_pubhub_incoming_data,buf);
		return;
	}

	/* we come here either after an error on read or when no more data are available */
	/* 1) remove the gdk tag */
	gdk_input_remove(fill_pubhub_handle_id);
	fill_pubhub_handle_id=-1;
	/* 2) close the com pipe */
	close(fill_pubhub_pipe);
	fill_pubhub_pipe=-1;
	/* 3) just to avoid problem, kill the hublist process */
	kill(fill_pubhub_procpid,SIGKILL);
	waitpid(fill_pubhub_procpid,NULL,0);
	fill_pubhub_procpid=-1;

	/* now, it's time to process the result */
	row=g_strsplit(fill_pubhub_incoming_data->str,"\r\n",0);	/* expand data (warning, this means a "\r" followed by a "\n", not ("\r" or "\n")) */
	gpa=g_ptr_array_new();
	gsc=g_string_chunk_new(128);

	SORT_SAVE(,fill_pubhub_hpc_gtm,PHC_RIGHT_ALIGN_COL)

	for(i=0;row[i]!=NULL;i++)
	{
		GtkTreeIter iter;
		char *ent[4];
		char *buf=row[i];
		char buf_cpy[51200];
		char *t;
		guint nb_users;

		strcpy(buf_cpy,buf);

		ent[0]=buf;

		t=strchr(buf,'|');
		if(t==NULL)
			continue;
		*t++='\0';

		ent[3]=t;
		t=strchr(t,'|');
		if(t==NULL)
			continue;
		*t++='\0';

		ent[2]=t;
		t=strchr(t,'|');
		if(t==NULL)
			continue;
		*t++='\0';

		ent[1]=t;
		t=strchr(t,'|');
		if(t==NULL)
			continue;
		*t++='\0';

		nb_users=atoi(ent[1]);

		if(nb_users>999999)	/* discard weird entries */
			continue;

		/* only add if a row with the same hub address does not exist */
		if(get_iter_from_string_nocase(fill_pubhub_hpc_gtm,PHC_HUB_ADDR_COL,ent[3],&iter)==FALSE)
		{
			gtk_list_store_append (fill_pubhub_hpc_gls, &iter);
			if(utf8_mode==TRUE)
			{
				gtk_list_store_set(fill_pubhub_hpc_gls,&iter,PHC_HUB_NAME_COL,ent[0],
																			PHC_NB_USERS_COL,nb_users,
																			PHC_DESC_COL,ent[2],
																			PHC_HUB_ADDR_COL,ent[3],
																			PHC_RIGHT_ALIGN_COL,1.0,
																			PHC_HUB_NAME,ent[0],
																			PHC_DESC,ent[2],
																			-1);
			}
			else
			{
				gchar *utf8_e0, *utf8_e2;

				utf8_e0=g_locale_to_utf8(ent[0],-1,NULL,NULL,NULL);
				utf8_e2=g_locale_to_utf8(ent[2],-1,NULL,NULL,NULL);
				gtk_list_store_set(fill_pubhub_hpc_gls,&iter,PHC_HUB_NAME_COL,utf8_e0,
																			PHC_NB_USERS_COL,nb_users,
																			PHC_DESC_COL,utf8_e2,
																			PHC_HUB_ADDR_COL,ent[3],
																			PHC_RIGHT_ALIGN_COL,1.0,
																			PHC_HUB_NAME,ent[0],
																			PHC_DESC,ent[2],
																			-1);
				g_free(utf8_e0);
				g_free(utf8_e2);
			}

			g_ptr_array_add(gpa,g_string_chunk_insert(gsc,buf_cpy));
		}
	}

	SORT_RESTORE(,fill_pubhub_hpc_gtm)

	if(gpa!=NULL)
	{
		update_seen_hub_list(gpa);
		g_ptr_array_free(gpa,TRUE);
	}

	if(gsc)
		g_string_chunk_free(gsc);

	if(row)
		g_strfreev(row);

	if(fill_pubhub_incoming_data)
	{
		g_string_free(fill_pubhub_incoming_data,TRUE);
		fill_pubhub_incoming_data=NULL;
	}
	if(gtk_notebook_get_current_page(GTK_NOTEBOOK(get_widget_by_widget_name(main_window,"main_notebook")))!=CONNECT_TAB)
		blink_on("connect_page");
	if(gtk_notebook_get_current_page(GTK_NOTEBOOK(get_widget_by_widget_name(main_window,"connect_notebook")))!=PUBLIC_HUB_TAB)
		blink_on("connect_public_label");

	play_the_sound_of_event(SOUND_HUBLIST);
}

/*****************************************************/
/* routine fixant le flag non bloquant sur un socket */
/*****************************************************/
void set_non_bloquant_sock(int socket_fd)
{
   long retval;

   retval=fcntl(socket_fd,F_GETFL);
   retval|= O_NONBLOCK;

   fcntl(socket_fd,F_SETFL,retval);
}


/**************************************************************************/
/* fill the clist named "hub_public_clist" with the list of existing hubs */
/**************************************************************************/
void fill_pub_hub_clist(int flag)
{
	int com_pipe[2];
	GtkWidget *fill_pubhub_hpc_wid=NULL;	/* it is the GtkWidget named "hub_public_clist" */

	if(fill_pubhub_handle_id!=-1)	/* update already in progress */
		return;

	fill_pubhub_hpc_wid=get_widget_by_widget_name(main_window,"hub_public_clist");
	if(fill_pubhub_hpc_wid==NULL)
		return;

	fill_pubhub_hpc_gls=GTK_LIST_STORE(fill_pubhub_hpc_gtm=gtk_tree_view_get_model(GTK_TREE_VIEW(fill_pubhub_hpc_wid)));

	if((flag==FALSE)&&(gtk_tree_model_is_not_empty(fill_pubhub_hpc_gtm)))
		return;

	if(pipe(com_pipe)==-1)
	{
		perror("fill_pub_hub_clist: pipe fail");
		return;
	}

	gtk_list_store_clear(fill_pubhub_hpc_gls);

	/* array must have been set to NULL at the end */
	if(fill_pubhub_incoming_data)
		g_string_free(fill_pubhub_incoming_data,TRUE);

	fill_pubhub_incoming_data=g_string_new("");	

	switch(fill_pubhub_procpid=fork())
	{
		case -1:
				perror("fill_pub_hub_clist: fork fail");
				/* close communication pipe */
				close(com_pipe[0]);
				close(com_pipe[1]);
	
				/* and free memory */
				if(fill_pubhub_incoming_data)
				{
					g_string_free(fill_pubhub_incoming_data,TRUE);
					fill_pubhub_incoming_data=NULL;
				}
				return;

		case 0:	/* it is the son, it will start hublist */
				close(com_pipe[0]);
				dup2(com_pipe[1],1);		/* connect the com_pipe to stdout */

				start_hublist();
				close(1);
				close(com_pipe[1]);
				_exit(0);

		default:
				close(com_pipe[1]);
				fill_pubhub_pipe=com_pipe[0];					/* pipe used to communicate with hub list (it is a read only pipe) */
				set_non_bloquant_sock(fill_pubhub_pipe);
				
				fill_pubhub_handle_id=gdk_input_add(fill_pubhub_pipe,GDK_INPUT_READ,async_fill_pub_hub_clist,NULL);
				break;
	}
}

/**************************************************/
/* build modele and view for the public hub clist */
/**************************************************/
void bmav4_public_hub_clist(void)
{
   GtkListStore *model;
   GtkWidget *view;
   GtkCellRenderer *rend;
   GtkTreeSelection *slc;

   model=gtk_list_store_new(NB_PHC_COL, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING,G_TYPE_STRING, G_TYPE_FLOAT, G_TYPE_STRING, G_TYPE_STRING);

   view=get_widget_by_widget_name(main_window,"hub_public_clist");

   gtk_tree_view_set_model(GTK_TREE_VIEW(view),GTK_TREE_MODEL(model));
   gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(view),TRUE);
#if 0
   gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW(view),TRUE);
#endif

   slc=gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
   gtk_tree_selection_set_mode(slc,GTK_SELECTION_MULTIPLE);

   rend=gtk_cell_renderer_textx_new();
	g_object_set(rend,"rowspacing",TRUE,NULL);
   MY_RS_gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(view),PHC_HUB_NAME_COL,_("Hubname"),rend,"text",PHC_HUB_NAME_COL,NULL);

   rend=gtk_cell_renderer_textx_new();
	g_object_set(rend,"rowspacing",TRUE,NULL);
   MY_RS_gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(view),PHC_NB_USERS_COL,_("#users"),rend,"text",PHC_NB_USERS_COL,"xalign",PHC_RIGHT_ALIGN_COL,NULL);

   rend=gtk_cell_renderer_textx_new();
	g_object_set(rend,"rowspacing",TRUE,NULL);
   MY_RS_gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(view),PHC_DESC_COL,_("Description"),rend,"text",PHC_DESC_COL,NULL);

   rend=gtk_cell_renderer_textx_new();
	g_object_set(rend,"rowspacing",TRUE,NULL);
   MY_RS_gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(view),PHC_HUB_ADDR_COL,_("Address"),rend,"text",PHC_HUB_ADDR_COL,NULL);

   /* The view now holds a reference.  We can get rid of our own
    * reference */
   g_object_unref (G_OBJECT (model));
}


